!! Copyright (C) 2004-2012 M. Oliveira, F. Nogueira, T. Cerqueira
!!
!! This program is free software; you can redistribute it and/or modify
!! it under the terms of the GNU General Public License as published by
!! the Free Software Foundation; either version 2, or (at your option)
!! any later version.
!!
!! This program is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU General Public License for more details.
!!
!! You should have received a copy of the GNU General Public License
!! along with this program; if not, write to the Free Software
!! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
!! 02111-1307, USA.
!!
!! $Id: atom_ps_inc.F90 767 2013-01-11 16:18:25Z micael $

  subroutine atom_create_ps(ps_atm, ae_atm, eigensolver)
    !-----------------------------------------------------------------------!
    ! Creates a pseudo-atom and the corresponding KB atom.                  !
    !-----------------------------------------------------------------------!
    type(atom_t),        intent(inout) :: ps_atm
    type(atom_t),        intent(in)    :: ae_atm
    type(eigensolver_t), intent(in)    :: eigensolver

    type input_data
      integer  :: n
      integer  :: l
      real(R8) :: j
      real(R8) :: rc
      integer  :: scheme
    end type input_data

    type channel
      type(states_batch_t) :: fold
      integer    :: scheme
      real(R8)   :: rc
      type(qn_t) :: qn
    end type channel

    integer :: i, j, k, n, ierr, n_cols, n_lines, n_channels, nlcc, file_format
    integer(POINTER_SIZE) :: blk
    real(R8) :: z_val, tol, min_ev, max_ev, ev, norm, slope, rmatch, nlcc_rc
    character(3) :: unbound_trick
    character(10) :: label
    real(R8), allocatable :: rc(:), core_density(:,:), ps_density(:,:), ps_v(:,:), vh(:,:), vxc(:,:), vxctau(:,:)
    type(input_data), allocatable :: id(:)
    type(qn_t), allocatable :: qn(:)
    type(state_t), pointer :: ae_state, state
    type(channel), allocatable :: channels(:)

    call push_sub("atom_create_ps")

    ASSERT(ae_atm%type == ATOM_AE)

    if (ae_atm%nspin == 2) then
      message(1) = "Generation of pseudopotentials from spin-polarized calculations"
      message(2) = "is not implemented!"
      message(3) = ""
      message(4) = "For most applications, spin-polarized calculations can be"
      message(5) = "performed using pseudopotentials generated for a spin-unpolarized"
      message(6) = "atom, as spin-polarization should come from the valence electrons."
      call write_fatal(6)
    end if

    ps_atm%type = ATOM_PS

    !Get information from all-electron atom
    !(except the atomic number, which will be obtained later)
    ps_atm%symbol = ae_atm%symbol
    ps_atm%theory_level = ae_atm%theory_level
    ps_atm%wave_eq = ae_atm%wave_eq
    ps_atm%nspin = ae_atm%nspin
    ps_atm%xc_model = ae_atm%xc_model
    ps_atm%m = ae_atm%m
    ps_atm%integrator_sp = ae_atm%integrator_sp
    ps_atm%integrator_dp = ae_atm%integrator_dp


                    !---Read input options---!

    !Print some information
    message(1) = ""
    message(2) = "Pseudo atom information:"
    call write_info(2)
    call write_info(2,unit=info_unit("pp"))

    !Get PP generation tolerance
    call oct_parse_float('PPCalculationTolerance', 1e-6_r8, tol)
    if (tol <= M_ZERO) then
      message(1) = "PPCalculationTolerance must be positive."
      call write_fatal(1)
    end if

    !Check how to deal with the unbound states
    call oct_parse_string('PPTrick', 'non', unbound_trick)

    !Open the orbitals block and check how many lines the it has
    ierr = oct_parse_block("PPComponents", blk)
    n_lines = oct_parse_block_n(blk)
    allocate(id(n_lines))

    !Read the input data
    do i = 1, n_lines
      n_cols = oct_parse_block_cols(blk, i-1)

      !Check for errors in the format
      if (n_cols < 4 .or. n_cols > 5 .or. &
           (ps_atm%wave_eq /= DIRAC .and. n_cols == 5)) then
        write(message(1),'("There is something wrong at line ",I2," of the PPComponents block.")') i
        call write_fatal(1)
      end if

      !Get main quantum number
      call oct_parse_block_int(blk, i-1, 0, id(i)%n)

      !Get angular momentum quantum number
      call oct_parse_block_int(blk, i-1, 1, id(i)%l)

      if (ps_atm%wave_eq == DIRAC .and. n_cols == 5) then
        !The third column is the total angular momentum quantum number
        call oct_parse_block_double(blk, i-1, 2, id(i)%j)
      else
        id(i)%j = M_ZERO
      end if

      !Get cutoff radius (it is always the column before the last)
      call oct_parse_block_double(blk, i-1, n_cols-2, id(i)%rc)
      id(i)%rc = id(i)%rc*units_in%length%factor

      !Get scheme (it is always the last column)
      call oct_parse_block_int(blk, i-1, n_cols-1, id(i)%scheme)

    end do
    call oct_parse_block_end(blk)

    !Check for errors in the input data 
    do i = 1, n_lines
      if (id(i)%rc < M_ZERO) then
        write(message(1),'("There is something wrong at line ",I2," of the PPComponents block.")') i
        message(2) = "The cutoff radius can not take negative values."
        call write_fatal(2)
      end if
      do j = i+1, n_lines
        if (id(i)%n == id(j)%n .and. id(i)%l == id(j)%l .and. id(i)%j == id(j)%j) then
          write(message(1),'("There is something wrong at line ",I2," of the PPComponents block.")') i
          message(2) = "There is at least another line with the same set of quantum numbers."
          call write_fatal(2)
        end if
      end do
      select case (id(i)%scheme)
      case (HAM,TM,MRPP,MTM)
      case (RTM,RMRPP)
        if (ps_atm%wave_eq /= DIRAC) then
          write(message(1),'("There is something wrong at line ",I2," of the PPComponents block.")') i
          message(2) = "All electron relativistic calculation is needed in order to use"
          message(3) = "the relativistic extensions to the TM or MRPP schemes."
          call write_fatal(2)
        end if
      case default
        write(message(1),'("There is something wrong at line ",I2," of the PPComponents block.")') i
        message(2) = "Illegal pseusopotential generation scheme."
        call write_fatal(2)
      end select

    end do


                    !---Initialize data structures---!

    !Allocate memory and copy some information
    n_channels = n_lines
    if (ps_atm%wave_eq == DIRAC) n_channels = n_channels + count(id%l /= 0 .and. id%j == M_ZERO)
    allocate(channels(n_channels))
    do i = 1, n_channels
      call states_batch_null(channels(i)%fold)
    end do
    k = 0
    do i = 1, n_lines
      k = k + 1
      if (ps_atm%wave_eq == DIRAC .and. id(i)%j == M_ZERO) then
        !Split "j=l +/- 1/2" states that were implicit in the input file
        channels(k)%qn = qn_init(id(i)%n, id(i)%l, M_ZERO, j=id(i)%l+M_HALF)
        channels(k)%rc = id(i)%rc
        channels(k)%scheme = id(i)%scheme
        if (id(i)%l /= M_ZERO) then
          k = k + 1
          channels(k)%qn = qn_init(id(i)%n, id(i)%l, M_ZERO, j=id(i)%l-M_HALF)
          channels(k)%rc = id(i)%rc
          channels(k)%scheme = id(i)%scheme
        end if
      else
        channels(k)%qn = qn_init(id(i)%n, id(i)%l, M_ZERO, j=id(i)%j)
        channels(k)%rc = id(i)%rc
        channels(k)%scheme = id(i)%scheme
      end if
    end do
    deallocate(id)

    !Create states batch and add states to the channels folds
    call states_batch_null(ps_atm%states)
    do i = 1, n_channels
      do n = 1, states_batch_size(ae_atm%states)
        ae_state => states_batch_get(ae_atm%states, n)
        if (state_qn(ae_state) == channels(i)%qn) then
          allocate(state)
          call state_null(state)
          state = ae_state
          call states_batch_add(ps_atm%states, state)
          call states_batch_add(channels(i)%fold, state)
          nullify(state)
          exit
        end if
        if (n == states_batch_size(ae_atm%states)) then
          write(message(1),'("Orbital ",A," not found in all-electron calculation.")') &
               trim(qn_label(channels(i)%qn))
          call write_fatal(1)
        end if
      end do
    end do

    !Sanity check
    if (n_channels /= states_batch_number_of_folds(ps_atm%states, .false.)) then
      write(message(1),'("There is something wrong at the PPComponents block.")')
      message(2) = "Only one orbital per angular momentum channel should be specified."
    end if

    !Check core radius
    do i = 1, n_channels
      if (channels(i)%rc == M_ZERO) then
        channels(i)%rc = state_default_rc(states_batch_get(channels(i)%fold, 1), channels(i)%scheme)
      end if
      if (channels(i)%rc == M_ZERO) then
        do j = 1, states_batch_size(ps_atm%states)
          channels(i)%rc = max(channels(i)%rc, state_default_rc(states_batch_get(ps_atm%states, j), channels(i)%scheme))
        end do
      end if
    end do

    !Do something about the unbound states
    if (unbound_trick == "fhi") then      
      max_ev = maxval(states_batch_eigenvalues(ae_atm%states), &
                      states_batch_eigenvalues(ae_atm%states) /= M_ZERO)
      do i = 1, states_batch_size(ps_atm%states)
        state => states_batch_get(ps_atm%states, i)
        if (state_eigenvalue(state) == M_ZERO) then
          call state_update(state, ps_atm%m, ps_atm%wave_eq, ae_atm%potential, ps_atm%integrator_dp, max_ev)
        end if
      end do
    end if

    !We will include in the valence space all the states that are higher in 
    !energy than the states used to generate the pseudopotentials
    do i = 1, n_channels
      min_ev = maxval(states_batch_eigenvalues(channels(i)%fold))
      do j = 1, states_batch_size(ae_atm%states)
        ae_state => states_batch_get(ae_atm%states, j)
        if (channels(i)%qn == state_qn(ae_state)) cycle
        if (state_eigenvalue(ae_state) > min_ev .and. qn_equal_fold(state_qn(ae_state), channels(i)%qn, .false.)) then
          allocate(state)
          call state_null(state)
          state = ae_state
          call states_batch_add(ps_atm%states, state)
          call states_batch_add(channels(i)%fold, state)
          nullify(state)
          exit
        end if      
      end do

      !Sanity check
      if ( (channels(i)%scheme == MRPP .or.  channels(i)%scheme == RMRPP) .and. &
           states_batch_size(channels(i)%fold) == 1) then
        write(message(1),'("There is something wrong at the PPComponents block.")')
        message(2) = "When using the MRPP scheme, at least two orbitals from the"
        message(3) = "same angular momentum channel are needed in the valence space,"
        message(4) = "but only one was found."
        call write_fatal(4)
      end if

    end do

    !Write information
    write(message(1),'(2X,"Wavefunction info:")')
    write(message(2),'(4X,"State   Occupation    Node radius   Peak radius   Default core radius")')
    n = 2
    do j = 1, n_channels
      call states_batch_sort(channels(j)%fold, SORT_EV)
      do i = 1, states_batch_size(channels(j)%fold)
        state => states_batch_get(channels(j)%fold, i)
        n = n + 1
        label = state_label(state)
        select case (len_trim(label))
        case (2)
          write(message(n),'(5X,A,5X,F7.2,7X,F8.3,6X,F8.3,10X,F8.3)') &
               trim(label), state_charge(state), &
               state_outermost_node(state), &
               state_outermost_peak(state), &
               state_default_rc(state, channels(j)%scheme)
        case (5)
          write(message(n),'(4X,A,3X,F7.2,7XF8.3,6X,F8.3,10X,F8.3)') &
               trim(label), state_charge(state), &
               state_outermost_node(state), &
               state_outermost_peak(state), &
               state_default_rc(state, channels(j)%scheme)
        end select
      end do
    end do
    message(n+1) = ""
    message(n+2) = "Pseudopotential Generation:"
    call write_info(n+2)
    call write_info(n+2,unit=info_unit("pp"))

    !Get core density
    allocate(core_density(ps_atm%m%np, 1))
    core_density = states_batch_density(ae_atm%states, 1, ps_atm%m) - &
                   states_batch_density(ps_atm%states, 1, ps_atm%m)


          !---Generate pseudopotentials---!

    allocate(ps_v(ps_atm%m%np, n_channels))
    ps_v = M_ZERO    
    do i = 1, n_channels
      call states_batch_psp_generation(channels(i)%fold, ps_atm%m, channels(i)%scheme, &
           ps_atm%wave_eq, tol, ae_atm%potential, ps_atm%integrator_sp, &
           ps_atm%integrator_dp, eigensolver, channels(i)%rc, ps_v(:,i))
    end do


        !---Unscreening and spin-average of the pseudopotentials---!


    !Non-linear core corrections
    call oct_parse_int('CoreCorrection', CC_NONE, nlcc)
    if (nlcc /= CC_NONE) then
      !Get rc
      allocate(ps_density(ps_atm%m%np, 1))
      ps_density = states_batch_density(ps_atm%states, 1, ps_atm%m)
      do i = ps_atm%m%np, 2, -1 
        if (ps_atm%m%r(i)**2*core_density(i, 1) > ps_atm%m%r(i)**2*ps_density(i, 1)) then
          nlcc_rc = ps_atm%m%r(i)
          exit
        end if
      end do
      deallocate(ps_density)

      if (nlcc == CC_FHI) then
        call oct_parse_float('CoreCorrectionCutoff', ps_atm%m%r(i), nlcc_rc)
      end if
    
      !Initialize the nlcc part of xc_model
      call xc_nlcc_init(ps_atm%xc_model, ps_atm%m, nlcc_rc, nlcc, core_density)
    end if
    deallocate(core_density)

    !Get screening
    allocate(vh(ps_atm%m%np, 1), vxc(ps_atm%m%np, 1), vxctau(ps_atm%m%np, 1))
    call hartree_potential(ps_atm%m, states_batch_charge_density(ps_atm%states, ps_atm%m), &
         states_batch_charge(ps_atm%states), vh=vh)
    call xc_potential(ps_atm%xc_model, ps_atm%m, ps_atm%states, 1, vxc=vxc, vxctau=vxctau)

    !Unscreen
    do i = 1, n_channels
      ps_v(:, i) = ps_v(:, i) - vh(:, 1) - vxc(:, 1)
    end do

    !Initialize ps_potential structure
    allocate(qn(n_channels))
    do i = 1, n_channels
      qn(i) = state_qn(states_batch_get(channels(i)%fold, 1))
    end do 
    call potential_init(ps_atm%potential, ps_atm%m, 1, vh + vxc, vxctau, n_channels, qn, ps_v)
    deallocate(ps_v, qn, vh, vxc, vxctau)


          !---Self-consistency test---!

    message(1) = ""
    message(2) = "Pseudopotentials Self-Consistency:"
    write(message(3),'(2X,"State  Eigenvalue [",A,"]    Norm Test   Slope Test")') units_out%energy%abbrev
    call write_info(3)
    write(info_unit("pp"),*)
    call write_info(3,unit=info_unit("pp"))
    do i = 1, n_channels
      do j = 1, min(states_batch_size(channels(i)%fold), 2)
        state => states_batch_get(channels(i)%fold, j)

        if (channels(i)%scheme == HAM) then
          rmatch = hamann_match_radius(ps_atm%m, channels(i)%rc)
        else
          rmatch = channels(i)%rc
        end if

        call state_test_consistency(ps_atm%m, ps_atm%wave_eq, eigensolver, ps_atm%integrator_sp, &
             ps_atm%integrator_dp, ae_atm%potential, ps_atm%potential, state, &
             rmatch, ev, norm, slope)

        label = state_label(state, full=.true.)
        select case (len_trim(label))
        case (2)
          write(message(1),'(4X,A,4X,F12.5,5X,F10.7,2X,F10.7)') trim(label), &
                                    ev/units_out%energy%factor, norm, slope
        case (5)
          write(message(1),'(2X,A,3X,F12.5,5X,F10.7,2X,F10.7)') trim(label), &
                                    ev/units_out%energy%factor, norm, slope
        end select
        call write_info(1)
        call write_info(1,unit=info_unit("pp"))
      end do
    end do

    !Get the correct atomic number
    z_val = states_batch_charge(ps_atm%states)
    ps_atm%z = ae_atm%z - states_batch_charge(ae_atm%states) + z_val

    !Start writing things to the pseudopotential file
    call oct_parse_int('PPOutputFileFormat', PSIO_UPF, file_format)
    call ps_io_init(ps_atm%m, ps_atm%wave_eq, channels(1)%scheme, ae_atm%z, &
         ps_atm%z, z_val, ps_atm%symbol, file_format)
    call potential_ps_io_set(ps_atm%potential)
    allocate(rc(states_batch_size(ps_atm%states)))
    do i = 1, states_batch_size(ps_atm%states)
      state => states_batch_get(ps_atm%states, i)
      do j = 1, n_channels
        if (qn_equal_fold(channels(j)%qn, state_qn(state), .false.)) then
          rc(i) = channels(j)%rc
          exit
        end if
      end do
    end do
    call states_batch_ps_io_set(ps_atm%states, ps_atm%m, rc)
    deallocate(rc)
    call xc_ps_io_set(ps_atm%xc_model)

    !Free memory
    do i = 1, n_channels
      call states_batch_end(channels(i)%fold)
    end do
    deallocate(channels)

    call pop_sub()
  end subroutine atom_create_ps

  subroutine atom_create_kb(kb_atm, ps_atm, eigensolver)
    !-----------------------------------------------------------------------!
    ! Generate the KB form of the pseudo-potentials.                        !
    !-----------------------------------------------------------------------!
    type(atom_t),        intent(inout) :: kb_atm
    type(atom_t),        intent(in)    :: ps_atm
    type(eigensolver_t), intent(in)    :: eigensolver

    integer :: l_local, i, p, k, n_kb
    real(R8) :: rcmax, v0, v1, v2, v3, r
    real(R8), allocatable :: vh(:,:), vxc(:,:), vxctau(:,:), vv(:), e(:), proj_f(:,:)
    type(qn_t) :: qn_loc
    type(qn_t), allocatable :: qn(:)
    type(potential_t) :: v_local
    type(states_batch_t), allocatable :: folds(:)
    type(mesh_t) :: m

    call push_sub("atom_create_kb")

    ASSERT(ps_atm%type == ATOM_PS)

    call potential_null(v_local)
    call mesh_null(m)

    kb_atm%type = ATOM_KB

    !Get information from pseudo-atom
    kb_atm%symbol = ps_atm%symbol
    kb_atm%theory_level = ps_atm%theory_level
    kb_atm%wave_eq = ps_atm%wave_eq
    kb_atm%nspin = ps_atm%nspin
    kb_atm%xc_model = ps_atm%xc_model
    kb_atm%m = ps_atm%m
    kb_atm%integrator_sp = ps_atm%integrator_sp
    kb_atm%integrator_dp = ps_atm%integrator_dp
    kb_atm%z = ps_atm%z

    !Shortcuts
    m = kb_atm%m

    !Write information
    message(1) = ""
    message(2) = "Kleinman & Bylander Atom"
    call write_info(2)
    call write_info(2, unit=info_unit("kb"))

    !Read input
    call oct_parse_int("Llocal", -1, l_local)
    if (l_local < -1) then
      message(1) = "Invalid value for Llocal."
      call write_fatal(1)
    end if
    if (l_local == -1) then
      write(message(1),'(2X,"Local potential is a Vanderbilt function")')
    else
      write(message(1),'(2X,"l-component used as local: l = ",I1)') l_local
    end if
    call write_info(1, 20)
    call write_info(1, unit=info_unit("kb"))

    !Divide states into folds. We need this, because there will be a KB projector per fold.
    n_kb = states_batch_number_of_folds(ps_atm%states, .false.)
    allocate(folds(n_kb))
    do i = 1, n_kb
      call states_batch_null(folds(i))
    end do
    call states_batch_split_folds(ps_atm%states, folds, .false.)

    ! Get local potential
    if (l_local == -1) then
      ! If llocal=-1, use a Vanderbilt function for the local component

      !Determine the radius at which all the pseudopotentials become equal
      allocate(vv(n_kb))
      do i = 1, m%np
        do k = 1, n_kb
          vv(k) = v(ps_atm%potential, m%r(i), state_qn(states_batch_get(folds(k), 1)), unscreened=.true.)
        end do
        if (all(vv == vv(1))) then
          rcmax = m%r(i-1)
          exit
        end if
      end do
      deallocate(vv)

      !Build Vanderbilt function
      v0 = -kb_atm%z*exp(M_ELEVEN/M_TWELVE)/rcmax
      v1 = -M_THREE/(M_TWO*rcmax**2)
      v2 = M_THREE/(M_FOUR*rcmax**4)
      v3 = -M_ONE/(M_SIX*rcmax**6)

      allocate(vv(m%np))
      do i = 1, m%np
        if (m%r(i) < rcmax) then
          vv(i) = v0*exp(v1*m%r(i)**2 + v2*m%r(i)**4 + v3*m%r(i)**6)
        else
          vv(i) = -kb_atm%z/m%r(i)
        end if
      end do

      !Write some information to the screen
      write(message(1),'(4X,3X,"z",5X,"rcmax",6X,"v0",10X,"v1",10X,"v2",10X,"v3")')
      write(message(2),'(4X,F6.2,2X,F6.2,2X,4(F10.6,2X))') kb_atm%z, rcmax, v0, v1, v2, v3
      call write_info(2)
      call write_info(2, unit=info_unit("kb"))

    else
      allocate(vv(m%np))
      qn_loc = QN_NULL
      qn_loc%l = l_local
      if (kb_atm%wave_eq == DIRAC .and. l_local /= 0) then
        qn_loc%j = l_local - M_HALF
        do k = 1, m%np
          vv(k) = v(ps_atm%potential, m%r(k), qn_loc, unscreened=.true.)
        end do
        qn_loc%j = l_local + M_HALF
        do k = 1, m%np
          vv(k) = M_ONE/(M_TWO*l_local + M_ONE)*(l_local*vv(k) + &
            (l_local + M_ONE)*v(ps_atm%potential, m%r(k), qn_loc, unscreened=.true.))
        end do
        
      else
        do k = 1, m%np
          vv(k) = v(ps_atm%potential, m%r(k), qn_loc, unscreened=.true.)
        end do
      end if
    end if

    !Get screening and initialize local potential
    allocate(vh(m%np, 1), vxc(m%np, 1), vxctau(m%np, 1))
    call hartree_potential(m, states_batch_charge_density(ps_atm%states, m), &
         states_batch_charge(ps_atm%states), vh=vh)
    call xc_potential(kb_atm%xc_model, m, ps_atm%states, 1, vxc=vxc, vxctau=vxctau)
    call potential_init(v_local, m, 1, vh + vxc, vxctau, vv)
    deallocate(vv, vh, vxc, vxctau)
    write(message(1),'(2X,"Non-local components:")')
    write(message(2),'(4X,"State   KB Energy [",A,"]   KB Cosine")') units_out%energy%abbrev
    call write_info(2)
    call write_info(2, unit=info_unit("kb"))

    !Get the projectors and the KB energy
    allocate(qn(n_kb), e(n_kb), proj_f(m%np, n_kb))
    do i = 1, n_kb
      call states_batch_sort(folds(i), SORT_EV)
      qn(i) = state_qn(states_batch_get(folds(i), 1))
      call state_kb_projector(m, v_local, ps_atm%potential, states_batch_get(folds(i), 1), &
                              e(i), proj_f(:,i))
      call states_batch_end(folds(i))
    end do
    deallocate(folds)

    !Initialize the KB potential
    call potential_init(kb_atm%potential, m, l_local, v_local, n_kb, qn, e, proj_f)

    !Test for ghost states
    message(1) = ""
    write(message(2),'(2X,"Ghost state analysis:")')
    call write_info(2)
    call write_info(2, unit=info_unit("kb"))
    do i = 1, states_batch_size(ps_atm%states)
      call state_test_ghost(m, kb_atm%wave_eq, kb_atm%integrator_sp, kb_atm%integrator_dp, &
                            eigensolver, kb_atm%potential, states_batch_get(ps_atm%states, i))
    end do

    !Print localization radii
    message(1) = ""
    write(message(2),'(2X,"Localization radii [",A,"]:")') trim(units_out%length%abbrev)

    r = M_ZERO
    do i = m%np, 1, -1
      if (abs(v(v_local, m%r(i), QN_NULL, unscreened=.true.) &
           + states_batch_charge(ps_atm%states)/m%r(i)) > 0.001) then
        r = m%r(i)
        exit
      end if
    end do
    write(message(3),'(4X,"Local:",F6.2)') r/units_out%length%factor

    k = 0
    do p = 1, n_kb
      if (qn(p)%l == l_local) cycle
      k = k + 1
      r = M_ZERO
      do i = m%np, 1, -1
        if (abs(proj_f(i, p)) > 0.001) then
          r = m%r(i)
          exit
        end if
      end do
      write(message(3+k),'(4X,"l = ",I1,":",F6.2)') qn(p)%l, r/units_out%length%factor
    end do
    call write_info(3+k)
    call write_info(3+k, unit=info_unit("kb"))

    !Deallocate memory
    deallocate(qn, e, proj_f)
    call mesh_end(m)

    !Finish writing things to the pseudopotential file
    call potential_ps_io_set(kb_atm%potential)
    call ps_io_save()

    call pop_sub()
  end subroutine atom_create_kb

  subroutine atom_test(ps_atm, eigensolver)
    !-----------------------------------------------------------------------!
    ! Test the pseudo-potentials tranferability.                            !
    !-----------------------------------------------------------------------!
    type(atom_t),        intent(inout) :: ps_atm
    type(eigensolver_t), intent(in)    :: eigensolver

    logical :: redo_scf
    integer :: tests, i
    integer :: wave_eq
    character(len=20) :: dir
    type(atom_t) :: ae_atm
    type(state_t), pointer :: state

    call push_sub("ps_atom_test")

    ASSERT(ps_atm%type == ATOM_PS)

    call oct_parse_logical('PPTestSCF', .false., redo_scf)
    if (redo_scf) then
      !Should we change the orbitals?
      if (oct_parse_isdef("PPTestOrbitals") /= 0) then
        do i = 1, states_batch_size(ps_atm%states)
          state => states_batch_get(ps_atm%states,i)
          call state_end(state)
        end do
        call states_batch_end(ps_atm%states)
      
        call atom_init_states_from_block(ps_atm, "PPTestOrbitals")
      end if

      !Are we using the same wave-equation?
      if (ps_atm%wave_eq /= SCHRODINGER) then
        call oct_parse_int('WaveEquation', ps_atm%wave_eq, wave_eq)
        if (ps_atm%wave_eq == SCALAR_REL .and. wave_eq == DIRAC) then
          message(1) = "It is not possible to solve the Dirac equation for pseudopotentials"
          message(2) = "generated from a scalar relativistic calculation."
          call write_fatal(2)
        end if
        ps_atm%wave_eq = wave_eq
      end if
      
      !Initialize integrator
      call integrator_init(ps_atm%integrator_sp, ps_atm%integrator_dp)

      !SCF cycle
      call atom_solve(ps_atm, eigensolver, info_unit("tests"))

      !Output wave-functions
      call atom_output(ps_atm, "tests")
    end if

    !Get the all electron atom
    call oct_parse_string("PPTestAEDir", "ae", dir)
    call atom_null(ae_atm)
    call atom_load(ae_atm, dir)

    !Now do the tests
    call oct_parse_int('PPTests', TEST_LD + TEST_DM, tests)

    if (iand(tests, TEST_LD).ne.0) then
      call atom_ld_test(ae_atm, ps_atm)
    end if

    if (iand(tests, TEST_DM).ne.0) then
      call atom_dm_test(ae_atm, ps_atm)
    end if
    
    !Ends
    call atom_end(ae_atm)

    call pop_sub()
  end subroutine atom_test

  subroutine atom_ld_test(ae_atm, ps_atm)
    !-----------------------------------------------------------------------!
    ! Test the pseudo-potentials tranferability by computing the            !
    ! logarithmic derivatives.                                              !
    !-----------------------------------------------------------------------!
    type(atom_t),       intent(in)    :: ae_atm
    type(atom_t),       intent(inout) :: ps_atm

    logical :: emin_is_def, emax_is_def
    integer :: i, n_folds
    real(R8) :: rc, rld, emin, emax, de
    real(R8), parameter :: covalent_radius(112) = &
         !   H     He    Li    Be     B     C     N     O     F    Ne    Na
         (/ 0.60, 1.76, 2.32, 1.70, 1.55, 1.46, 1.42, 1.38, 1.36, 1.34, 2.91, &
         !   Mg    Al    Si     P     S    Cl    Ar     K    Ca    Sc    Ti   
            2.57, 2.23, 2.10, 2.00, 1.93, 1.87, 1.85, 3.84, 3.29, 2.72, 2.49, &
         !    V    Cr    Mn    Fe    Co    Ni    Cu    Zn    Ga    Ge    As
            2.31, 2.23, 2.21, 2.21, 2.19, 2.17, 2.21, 2.36, 2.38, 2.31, 2.27, &
         !   Se    Br    Kr    Rb    Sr     Y    Zr    Nb    Mo    Tc    Ru
            2.19, 2.15, 2.12, 4.08, 3.61, 3.06, 2.74, 2.53, 2.46, 2.40, 2.36, &
         !   Rh    Pd    Ag    Cd    In    Sn    Sb    Te     I    Xe    Cs
            2.36, 2.42, 2.53, 2.80, 2.72, 2.66, 2.66, 2.57, 2.51, 2.48, 4.44, &
         !   Ba    La    Ce    Pr    Nd    Pm    Sm    Eu    Gd    Tb    Dy
            3.74, 3.19, 3.12, 3.12, 3.10, 3.08, 3.06, 3.50, 3.04, 3.00, 3.00, &
         !   Ho    Er    Tm    Yb    Lu    Hf    Ta     W    Re    Os    Ir
            2.99, 2.97, 2.95, 3.29, 2.95, 2.72, 2.53, 2.46, 2.42, 2.38, 2.40, &
         !   Pt    Au    Hg    Tl    Pb    Bi    Po    At    Rn    Fr    Ra
            2.46, 2.53, 2.82, 2.80, 2.78, 2.76, 2.76, 2.74, 2.74, 0.00, 0.00, &
         !   Ac    Th    Pa     U    Np    Pu    Am    Cm    Bk    Cf    Es
            0.00, 3.12, 0.00, 2.68, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, &
         !   Fm    Md    No    Lr    Rf    Db    Sg    Bh    Hs    Mt    Uun
            0.00, 0.00, 0.00, 2.68, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, &
         !   Uuu   Uub
            0.00, 0.00 /)
    type(states_batch_t), allocatable :: folds(:)

    call push_sub("ps_atom_ld_test")

    ASSERT(ae_atm%type == ATOM_AE .and. ps_atm%type == ATOM_PS)

    message(1) = ""
    message(2) = "Logarithmic Derivatives:"
    call write_info(2)
    call write_info(2,unit=info_unit("tests"))

    !Read the input options
    rc = covalent_radius(int(ae_atm%z))/units_in%length%factor
    call oct_parse_float("LogDerivativeRadius", rc, rld)
    if (rld <= M_ZERO) then
      write(message(1), '(A,F14.6,A)') "Input: '",rld,"' is not a valid LogDerivativeRadius"
      message(2) = '(LogDerivativeRadius > 0.0)'
      call write_fatal(2)
    end if
    rld = rld*units_in%length%factor

    emin_is_def = oct_parse_isdef("LogDerivativeEnergyMin") == 1
    if (emin_is_def) then
      call oct_parse_float("LogDerivativeEnergyMin", M_TWO, emin)
      emin = emin*units_in%energy%factor
    end if

    emax_is_def = oct_parse_isdef("LogDerivativeEnergyMax") == 1
    if (emax_is_def) then
      call oct_parse_float("LogDerivativeEnergyMax", M_TWO, emax)
      emax = emax*units_in%energy%factor
    end if

    call oct_parse_float("LogDerivativeEnergyStep", M_ZERO, de)
    if (de < M_ZERO) then
      message(1) = "LogDerivativeEnergyStep must be greater then zero"
    end if
    de = de*units_in%energy%factor

    !Output input info to the screen and to file tests/info
    write(message(1), '(2X,"Diagnostic radius: ",F6.3,1X,A)') &
         rld/units_out%length%factor, trim(units_out%length%abbrev)
    if (de == M_ZERO) then
      write(message(2), '(2X,"Energy step:    Adaptive")')
    else
      write(message(2), '(2X,"Energy step:    ",F6.3,1X,A)') &
           de/units_out%energy%factor, trim(units_out%energy%abbrev)
    end if
    call write_info(2)
    call write_info(2,unit=info_unit("tests"))

    !Split states into folds. The logarithmic derivatives will be computed for each fold
    n_folds = states_batch_number_of_folds(ps_atm%states, .false.)
    allocate(folds(n_folds))
    do i = 1, n_folds
      call states_batch_null(folds(i))
    end do
    call states_batch_split_folds(ps_atm%states, folds, .false.)

    !Perform test for each fold
    do i = 1, n_folds
      if (emin_is_def .and. emax_is_def) then
        call states_batch_ld_test(folds(i), ae_atm%m, ae_atm%potential, &
             ps_atm%potential, ps_atm%integrator_dp, rld, de, emin, emax)
      elseif (emin_is_def .and. .not.emax_is_def) then
        call states_batch_ld_test(folds(i), ae_atm%m, ae_atm%potential, &
             ps_atm%potential, ps_atm%integrator_dp, rld, de, emin=emin)
      elseif (.not. emin_is_def .and. emax_is_def) then
        call states_batch_ld_test(folds(i), ae_atm%m, ae_atm%potential, &
             ps_atm%potential, ps_atm%integrator_dp, rld, de, emax=emax)
      else
        call states_batch_ld_test(folds(i), ae_atm%m, ae_atm%potential, &
             ps_atm%potential, ps_atm%integrator_dp, rld, de)
      end if
    end do

    !Free memory
    do i = 1, n_folds
      call states_batch_end(folds(i))
    end do
    deallocate(folds)

    call pop_sub()
  end subroutine atom_ld_test

  subroutine atom_dm_test(ae_atm, ps_atm)
    !-----------------------------------------------------------------------!
    ! Test the pseudo-potentials tranferability by computing the            !
    ! dipole matrix elements.                                               !
    !-----------------------------------------------------------------------!
    type(atom_t), intent(in) :: ae_atm, ps_atm

    integer :: i, f, k
    real(R8) :: ae_mif, ps_mif
    character(len=10) :: label_i, label_f
    type(state_t), pointer :: ae_state_i, ae_state_f, ps_state_i, ps_state_f

    call push_sub("ps_atom_dm_test")

    ASSERT(ae_atm%type == ATOM_AE .and. ps_atm%type == ATOM_PS)

    message(1) = ""
    message(2) = "Dipole Matrix Elements:"
    write(message(3),'(6X,"States",11X,"AE",9X,"PS")')
    call write_info(3)
    call write_info(3,unit=info_unit("tests"))

    !Loop over i and f states
    i_loop: do i = 1, states_batch_size(ps_atm%states) - 1
      ps_state_i => states_batch_get(ps_atm%states, i)

      !Get ae state i
      do k = 1, states_batch_size(ae_atm%states)
        ae_state_i => states_batch_get(ae_atm%states, k)
        if (ae_state_i == ps_state_i) exit
      end do

      f_loop: do f = i + 1, states_batch_size(ps_atm%states)
        ps_state_f => states_batch_get(ps_atm%states, f)

        !Get ae state f
        do k = 1, states_batch_size(ae_atm%states)
          ae_state_f => states_batch_get(ae_atm%states, k)
          if (ae_state_f == ps_state_f) exit
        end do

        !Compute matrix elements
        ae_mif = state_dipole_matrix_element(ps_atm%m, ae_state_i, ae_state_f)
        ps_mif = state_dipole_matrix_element(ps_atm%m, ps_state_i, ps_state_f)

        !Write results to screen
        label_i = state_label(ae_state_i, full=.true.)
        label_f = state_label(ae_state_f, full=.true.)
        select case (len_trim(label_i))
        case (2)
          write(message(1),'(5X,A," -- ",A,7X,F7.4,4X,F7.4)') &
                trim(label_i), trim(label_f), ae_mif, ps_mif
        case (5)
           write(message(1),'(2X,A," -- ",A,4X,F7.4,4X,F7.4)') &
                trim(label_i), trim(label_f), ae_mif, ps_mif
        end select
        call write_info(1)
        call write_info(1,unit=info_unit("tests"))
        
      end do f_loop

    end do i_loop

    call pop_sub()
  end subroutine atom_dm_test
